from pathlib import Path, PurePosixPath
from rich import print
import subprocess
from sys import exit
from tabulate import tabulate
from typing import List, Dict
from jira import JIRA, Issue
from yaml import safe_load
import re
from dataclasses import dataclass
from datetime import datetime
from powerjira_toolchain.params import *

jira = JIRA(server=domain, basic_auth=(user_name, token))


def errorMessage(message:str) -> None:
  '''generic error messageing'''
  print(f'[red][bold]Error[/bold]. {message}[/red]')
  exit(1)


def verifyPath(path:str) -> None:
  '''check if input and output paths are valid, exits if not'''
  path_object = Path(path)
  format = PurePosixPath(path).suffix.lower()[1:]
  if not path_object.exists():
    errorMessage(f'Path {path} does not exist. Run `init` command.')
    exit(1)
  elif not path_object.is_file():
    errorMessage(f'Path {path} is not a file.  Run `init` command.')
    exit(1)
  elif format not in ['yml', 'txt']:
    errorMessage(f'Input format "{format}" not supported, must be one of: yml, .txt\nRun `init` command.')
    exit(1)


def readConfig(config_path:str) -> Dict[str,any]:
  '''reads config file and returns config as dict'''
  with open(config_path, 'r') as raw_config:
    return safe_load(raw_config)


def ensureTemplateTree(destination_path:str) -> None:
  """
  Ensures the existence of the following file tree, creating when needed:
    <powerjira_directory>/
      - ticket.yml
      - summary.txt
      - description.txt
  """
  destination = Path(destination_path).expanduser().resolve()
  destination.mkdir(parents=True, exist_ok=True)

  for filename,content in templates.items():
    dest_item = destination / filename
    if not dest_item.exists():
      with open(dest_item, 'w') as f:
        f.write(content)
        print(f"Created file {destination}/{filename}.")


def openPowerJiraDirectory(editor:str) -> None:
  '''opens powerjira directory in editor'''
  match editor:
    case 'vscode':
      subprocess.run(f'code -n {powerjira_directory}', shell=True, capture_output=True, text=True, executable=powerjira_shell)
    case _:
      errorMessage(f'Editor {editor} not supported.')


def formatTicketString(ticket:Issue, branch_name:str, parent_branch:str, style:str=result_table_style) -> str:
  '''builds string per ticket info to print'''
  #───GLANCE───────────────────
  if len(ticket.fields.summary) < ticket_excerpt_length:
    formatted_summary = ticket.fields.summary
  else:
    formatted_summary = ticket.fields.summary[:ticket_excerpt_length-3] + '...'
  if ticket.fields.description:
    if len(ticket.fields.description) < ticket_excerpt_length:
      formatted_description = ticket.fields.description
    else:
      formatted_description = ticket.fields.description[:ticket_excerpt_length-3] + '...'
  else:
    formatted_description = "(no description provided)"

  payload = '\n[bold green]GLANCE[/bold green]\n'
  payload += f'{ticket.fields.reporter} 👉 {ticket.fields.assignee}'
  payload += f'\n[bold italic]───Summary[/bold italic]\n{formatted_summary}'
  payload += f'\n[bold italic]───Description[/bold italic]\n{formatted_description}\n'

  #───TICKET───────────────────
  created_date_formatted = datetime.strptime(ticket.fields.created, "%Y-%m-%dT%H:%M:%S.%f%z").strftime('%Y-%m-%d %H:%M:%S')
  info_table = tabulate([
    ['Key', ticket.key],
    ['Url', f'{domain}/browse/{ticket.key}'],
    ['Type', ticket.fields.issuetype.name],
    ['Created', created_date_formatted],
    ['Status', ticket.fields.status.name],
    ['Priority', ticket.fields.priority.name if ticket.fields.priority else 'None']
  ], tablefmt=style)
  payload += f'\n[bold green]TICKET[/bold green]\n'
  payload += f'{info_table}\n'

  #───GLANCE───────────────────
  payload += f'\n[bold green]GIT COMMANDS[/bold green]\n'
  git_table = tabulate([
    ['Create Branch', f'git checkout -b {branch_name}'],
    ['GitLab MR', f'git push origin {branch_name} -o merge_request.create -o merge_request.target={parent_branch}'],
    ['GitHub PR', f"gh pr create --title 'TITLE' --body 'BODY' --base '{parent_branch}' --head '{branch_name}'"]
  ], tablefmt=style)
  payload += f'{git_table}\n'

  return payload


def buildTable(data:List[any], title:str, headers:List[str]=None, style:str=result_table_style) -> str:
  '''assembles table, with all the fixings'''
  table = tabulate(data, headers=headers, tablefmt=style)
  return f"\n[bold green]{title}[/bold green]\n{table}\n"


def getUserID(search_name:str, session_object=jira) -> str:
  '''Get USER ID
  fetches a user's user-id from jira, can be:
    - email (iyep@halcyon.ai)
    - email without domain (iyep)
    - full name (Isaac Yep)
    - first name and last initial (Isaac Y)
  '''
  user_search_result_list = session_object.search_users(query=search_name)
  if user_search_result_list:
    return user_search_result_list[0].accountId
  else:
    print(f"No users found by search term `{search_name}`")
    exit(1)


def substituteDynamicParams(input_string:str, params_dict:Dict[str, any]) -> str:
  """Substitute Dynamic Parameters
  Replaces substrings enclosed in angle brackets with corresponding values from a dictionary.

  Args:
    input_string (str): The input string containing substrings enclosed in angle brackets.
    params_dict (dict): A dictionary mapping parameter names to their replacement values.

  Returns:
    str: The input string with parameters substituted.
  """
  def replaceMatch(match):
    key = match.group(1)
    return params_dict.get(key, match.group(0))  # Replace if key exists, else keep original

  pattern = r"<(.*?)>"
  return re.sub(pattern, replaceMatch, input_string)


@dataclass
class regexCheck:
  """Regex checker match-case usage"""
  string: str
  def __eq__(self, other: str | re.Pattern):
    if isinstance(other, str):
      other = re.compile(other)
    assert isinstance(other, re.Pattern)
    return other.fullmatch(self.string) is not None



from rich import print
from rich.pretty import Pretty
from sys import exit
import typer
from jira import JIRAError
from powerjira_toolchain.params import *
from powerjira_toolchain.utils import *

app = typer.Typer(help="Jira Tickets without the guff")


@app.command()
def init():
  '''Creates the powerjira directory structure, or populates missing files. Opens in editor'''
  ensureTemplateTree(powerjira_directory)
  openPowerJiraDirectory(powerjira_editor)


@app.command()
def goto():
  '''Opens powerjira directory in edtitor'''
  openPowerJiraDirectory(powerjira_editor)


@app.command()
def make(dry_run:bool=False):
  '''Builds ticket per <powerjira_directory>/ticket.yml'''
  # file guards
  verifyPath(f'{powerjira_directory}/ticket.yml')
  verifyPath(f'{powerjira_directory}/summary.txt')
  verifyPath(f'{powerjira_directory}/description.txt')

  # read config
  config = readConfig(powerjira_directory + '/ticket.yml')
  try:
    reporter = config['reporter']
    assignee = config['assignee']
    init_status = config['init_status']
    project = config['project'].upper()
    priority = config['priority'].title()
    issue_type = config['issue_type'].title()
    parent_epic = config['parent_epic']
    parent_branch = config['parent_branch']
    branch_name_params = config['branch_name_params']
    branch_naming_convention = config['branch_naming_convention']
  except KeyError as ke:
    errorMessage(f'Missing config key: {ke}')

  # guard for <ticket_key> in branch_naming_convention
  if '<ticket_key>' not in branch_naming_convention:
    errorMessage('Branch naming convention must include <ticket_key>.')

  # build payload
  with open(f'{powerjira_directory}/summary.txt', 'r') as summary, open(f'{powerjira_directory}/description.txt', 'r') as description:
    summary = summary.read().rstrip('\n')
    description = description.read().rstrip('\n')
  ticket_blueprint = {
      'reporter': {'accountId': getUserID(reporter)},
      'assignee': {'accountId': getUserID(assignee)},
      'project': {'key': project},
      'issuetype': {'name': issue_type},
      'summary': summary,
      'description': description
    }
  if issue_type.lower() == 'task' :
    ticket_blueprint['priority'] = {'name': priority}
    if parent_epic:
      ticket_blueprint['parent'] = {'key': parent_epic}
  elif issue_type.lower() == 'epic' and parent_epic:
    errorMessage('Can not assign epic as child to another epic.')

  # make or dry-run
  if dry_run:
    print('ℹ️ Dry Run: Would have created:')
    print(Pretty(ticket_blueprint, expand_all=True))
    exit(0)
  else:
    new_issue = jira.create_issue(fields=ticket_blueprint)
    status_id = jira.find_transitionid_by_name(new_issue.key, init_status)
    jira.transition_issue(new_issue.key, transition=status_id)
    branch_name = substituteDynamicParams(branch_naming_convention.replace('<ticket_key>', new_issue.key), branch_name_params)
    print(formatTicketString(jira.issue(new_issue.key), branch_name=branch_name,  parent_branch=parent_branch))
    exit(0)


@app.command()
def fetch(target:str):
  '''Provide ticket key, or a status for all your related open tickets'''
  match regexCheck(target):
    case r"^.+-\d+$": # ticket key
      verifyPath(f'{powerjira_directory}/ticket.yml')
      config = readConfig(powerjira_directory + '/ticket.yml')
      try:
        parent_branch = config['parent_branch']
        branch_name_params = config['branch_name_params']
        branch_naming_convention = config['branch_naming_convention']
      except KeyError as ke:
        errorMessage(f'Missing config key: {ke}')
      try:
        issue = jira.issue(target)
        branch_name = substituteDynamicParams(branch_naming_convention.replace('<ticket_key>', issue.key), branch_name_params)
        print(formatTicketString(issue, branch_name=branch_name, parent_branch=parent_branch))
      except JIRAError as e:
        errorMessage(f'Ticket [bold]{target}[/bold] not found.')
      exit(0)

    case _: # assume status
      try:
        issues = jira.search_issues(f'status = "{target}" AND assignee = currentUser()')
        print(buildTable([[t.key, t.fields.summary] for t in issues], title=f"My '{target}'", headers=['Ticket Key', 'Summary (title)']))
      except JIRAError as e:
        errorMessage(f'Status [bold]{target}[/bold] not found.')
      exit(0)


@app.command()
def watched(action:str):
  '''[list, done] Manage watched issues of status DONE'''
  watched_issues = jira.search_issues('watcher = currentUser() AND resolution = Done')

  title = {
    'prune': 'Un-Followed Issues',
    'list': 'Watched Issues'
  }[action]
  payload = []
  
  for issue in watched_issues:
    match action:
      case 'prune':
        resolution = str(issue.fields.resolution)
        if resolution.lower() == 'done':
          jira.remove_watcher(issue.key, user_name)
          payload.append([issue.key, issue.fields.summary])
      case 'list':
        payload.append([issue.key, issue.fields.summary])

  print(buildTable(payload, title=title, headers=['Ticket Key', 'Summary (title)']))
  exit(0)



#!/usr/bin/env python

'''README

Usage:
  ./powerjira.py init
  ./powerjira.py goto
  ./powerjira.py fetch <target>
  ./powerjira.py make
  ./powerjira.py watched list
  ./powerjira.py watched prune
'''

from powerjira_toolchain.commands import app

if __name__ == "__main__":
  app()



from os import getenv, path
from yaml import safe_load

# environment
user_name = getenv('JIRA_USERNAME')
token = getenv('JIRA_TOKEN')
domain = getenv('JIRA_DOMAIN')
home_dir = getenv('HOME')

# static
ticket_excerpt_length = 60
powerjira_directory = f'{home_dir}/.sleepyconfig/powerjira'
powerjira_editor = 'vscode'

# init templates
templates = {
  'ticket.yml': '''# pj --help

#───TICKET───────────────────
reporter: jack
assignee: jill

init_status: to do # new, to do, in progress, done, backlog

project: sales # sales, marketing, frontend
priority: medium # low, medium, high (ignored if epic)
issue_type: task # task, epic
parent_epic: '' # leave empty for standalone task or epic


#───GIT──────────────────────
parent_branch: main
# ℹ️ naming convention must include <ticket_key>
branch_naming_convention: feature/<ticket_key>_<branch_suffix>
branch_name_params: # custom branch name parameters
  branch_suffix: implement_thing''',

  'summary.txt': '''This is the ticket's title, which jira calls the "summary"''',

  'description.txt': '''This is the ticket's description''',
}


#───CONFIG FILE──────────────
global_config_path = f'{home_dir}/.sleepyconfig/params.yml'
config_file_exists = path.exists(global_config_path)

def resolveValue(default:any, config_key:str) -> any:
  '''returns config value if exists, else default'''
  if not config_file_exists:
    return default
  with open(global_config_path, 'r') as f:
    raw_config = safe_load(f)
    if config_key not in raw_config:
      return default
    return raw_config[config_key]

## defaults
default_powerjira_table_style = 'rounded_outline'
default_shell = '/bin/zsh'
## config file
result_table_style = resolveValue(default_powerjira_table_style, 'pj_table_style')
powerjira_shell = resolveValue(default_shell, 'subprocess_shell')